@@ -31,6 +31,8 @@ import org.apache.ratis.protocol.Message;
 import org.apache.ratis.protocol.RaftClientReply;
 import org.apache.ratis.protocol.RaftClientRequest;
 import org.apache.ratis.protocol.RaftPeerId;
+import org.apache.ratis.protocol.exceptions.AlreadyClosedException;
+import org.apache.ratis.protocol.exceptions.AlreadyExistsException;
 import org.apache.ratis.protocol.exceptions.GroupMismatchException;
 import org.apache.ratis.protocol.exceptions.LeaderSteppingDownException;
 import org.apache.ratis.protocol.exceptions.StateMachineException;
@@ -99,7 +101,7 @@ class BlockingImpl implements BlockingApi {
           return client.handleReply(request, reply);
         }
       } catch (GroupMismatchException | StateMachineException | TransferLeadershipException |
-          LeaderSteppingDownException e) {
+          LeaderSteppingDownException | AlreadyClosedException | AlreadyExistsException e) {
         throw e;
       } catch (IOException e) {
         ioe = e;
@@ -123,7 +125,7 @@ class BlockingImpl implements BlockingApi {
     }
   }
 
-  RaftClientReply sendRequest(RaftClientRequest request) throws IOException {
+  private RaftClientReply sendRequest(RaftClientRequest request) throws IOException {
     LOG.debug("{}: send {}", client.getId(), request);
     RaftClientReply reply;
     try {
