@@ -21,7 +21,10 @@ import org.apache.ratis.client.DataStreamClient;
 import org.apache.ratis.client.RaftClient;
 import org.apache.ratis.client.RaftClientRpc;
 import org.apache.ratis.client.api.DataStreamApi;
+import org.apache.ratis.client.api.LeaderElectionManagementApi;
+import org.apache.ratis.client.api.SnapshotManagementApi;
 import org.apache.ratis.client.retry.ClientRetryEvent;
+import org.apache.ratis.conf.Parameters;
 import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.proto.RaftProtos.SlidingWindowEntry;
 import org.apache.ratis.protocol.ClientId;
@@ -139,10 +142,14 @@ public final class RaftClientImpl implements RaftClient {
   private final MemoizedSupplier<DataStreamApi> dataStreamApi;
 
   private final Supplier<AdminImpl> adminApi;
-  private final ConcurrentMap<RaftPeerId, GroupManagementImpl> groupManagmenets = new ConcurrentHashMap<>();
+  private final ConcurrentMap<RaftPeerId, GroupManagementImpl> groupManagement = new ConcurrentHashMap<>();
+  private final ConcurrentMap<RaftPeerId, SnapshotManagementApi> snapshotManagement = new ConcurrentHashMap<>();
+  private final ConcurrentMap<RaftPeerId, LeaderElectionManagementApi>
+      leaderElectionManagement = new ConcurrentHashMap<>();
 
+  @SuppressWarnings("checkstyle:ParameterNumber")
   RaftClientImpl(ClientId clientId, RaftGroup group, RaftPeerId leaderId, RaftPeer primaryDataStreamServer,
-      RaftClientRpc clientRpc, RaftProperties properties, RetryPolicy retryPolicy) {
+      RaftClientRpc clientRpc, RetryPolicy retryPolicy, RaftProperties properties, Parameters parameters) {
     this.clientId = clientId;
     this.peers.set(group.getPeers());
     this.groupId = group.getGroupId();
@@ -168,6 +175,7 @@ public final class RaftClientImpl implements RaftClient {
         .setRaftGroupId(groupId)
         .setDataStreamServer(primaryDataStreamServer)
         .setProperties(properties)
+        .setParameters(parameters)
         .build());
     this.adminApi = JavaUtils.memoize(() -> new AdminImpl(this));
   }
@@ -240,7 +248,22 @@ public final class RaftClientImpl implements RaftClient {
 
   @Override
   public GroupManagementImpl getGroupManagementApi(RaftPeerId server) {
-    return groupManagmenets.computeIfAbsent(server, id -> new GroupManagementImpl(id, this));
+    return groupManagement.computeIfAbsent(server, id -> new GroupManagementImpl(id, this));
+  }
+
+  @Override
+  public SnapshotManagementApi getSnapshotManagementApi() {
+    return JavaUtils.memoize(() -> new SnapshotManagementImpl(null, this)).get();
+  }
+
+  @Override
+  public SnapshotManagementApi getSnapshotManagementApi(RaftPeerId server) {
+    return snapshotManagement.computeIfAbsent(server, id -> new SnapshotManagementImpl(id, this));
+  }
+
+  @Override
+  public LeaderElectionManagementApi getLeaderElectionManagementApi(RaftPeerId server) {
+    return leaderElectionManagement.computeIfAbsent(server, id -> new LeaderElectionManagementImpl(id, this));
   }
 
   @Override
