@@ -27,6 +27,10 @@ public interface AdminProtocol {
 
   RaftClientReply groupManagement(GroupManagementRequest request) throws IOException;
 
+  RaftClientReply snapshotManagement(SnapshotManagementRequest request) throws IOException;
+
+  RaftClientReply leaderElectionManagement(LeaderElectionManagementRequest request) throws IOException;
+
   RaftClientReply setConfiguration(SetConfigurationRequest request) throws IOException;
 
   RaftClientReply transferLeadership(TransferLeadershipRequest request) throws IOException;
