@@ -27,6 +27,7 @@ import org.apache.ratis.protocol.Message;
 import org.apache.ratis.protocol.RaftClientRequest;
 import org.apache.ratis.protocol.RaftGroupId;
 import org.apache.ratis.protocol.RaftGroupMemberId;
+import org.apache.ratis.protocol.RaftPeer;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.protocol.TermIndex;
@@ -35,11 +36,13 @@ import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.thirdparty.com.google.protobuf.InvalidProtocolBufferException;
 import org.apache.ratis.util.JavaUtils;
 import org.apache.ratis.util.LifeCycle;
+import org.apache.ratis.util.ReferenceCountedObject;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
 import java.io.Closeable;
 import java.io.IOException;
+import java.nio.ByteBuffer;
 import java.nio.channels.WritableByteChannel;
 import java.util.Collection;
 import java.util.concurrent.CompletableFuture;
@@ -72,9 +75,8 @@ public interface StateMachine extends Closeable {
      *
      * @return a future for the read task.
      */
-    @SuppressFBWarnings("NP_NULL_PARAM_DEREF")
     default CompletableFuture<ByteString> read(LogEntryProto entry) {
-      return CompletableFuture.completedFuture(null);
+      throw new UnsupportedOperationException("This method is NOT supported.");
     }
 
     /**
@@ -201,11 +203,16 @@ public interface StateMachine extends Closeable {
      * Notify the {@link StateMachine} that the given follower is slow.
      * This notification is based on "raft.server.rpc.slowness.timeout".
      *
-     * @param roleInfoProto information about the current node role and rpc delay information
+     * @param leaderInfo information about the current node role and rpc delay information
+     * @param slowFollower The follower being slow.
      *
      * @see org.apache.ratis.server.RaftServerConfigKeys.Rpc#SLOWNESS_TIMEOUT_KEY
      */
-    default void notifyFollowerSlowness(RoleInfoProto roleInfoProto) {}
+    default void notifyFollowerSlowness(RoleInfoProto leaderInfo, RaftPeer slowFollower) {}
+
+    /** @deprecated Use {@link #notifyFollowerSlowness(RoleInfoProto, RaftPeer)}. */
+    @Deprecated
+    default void notifyFollowerSlowness(RoleInfoProto leaderInfo) {}
 
     /**
      * Notify {@link StateMachine} that this server is no longer the leader.
@@ -250,6 +257,40 @@ public interface StateMachine extends Closeable {
    * For write state machine data.
    */
   interface DataChannel extends WritableByteChannel {
+    /**
+     * This method is the same as {@link WritableByteChannel#write(ByteBuffer)}.
+     *
+     * If the implementation has overridden {@link #write(ReferenceCountedObject)},
+     * then it does not have to override this method.
+     */
+    @Override
+    default int write(ByteBuffer buffer) throws IOException {
+      throw new UnsupportedOperationException();
+    }
+
+    /**
+     * Similar to {@link #write(ByteBuffer)}
+     * except that the parameter is a {@link ReferenceCountedObject}.
+     *
+     * This is an optional method.
+     * The default implementation is the same as write(referenceCountedBuffer.get()).
+     *
+     * The implementation may choose to override this method in order to retain the buffer for later use.
+     *
+     * - If the buffer is retained, it must be released afterward.
+     *   Otherwise, the buffer will not be returned, and it will cause a memory leak.
+     *
+     * - If the buffer is retained multiple times, it must be released the same number of time.
+     *
+     * - It is safe to access the buffer before this method returns with or without retaining it.
+     *
+     * - If the buffer is not retained but is accessed after this method returns,
+     *   the content of the buffer could possibly be changed unexpectedly, and it will cause data corruption.
+     */
+    default int write(ReferenceCountedObject<ByteBuffer> referenceCountedBuffer) throws IOException {
+      return write(referenceCountedBuffer.get());
+    }
+
     /**
      * Similar to {@link java.nio.channels.FileChannel#force(boolean)},
      * the underlying implementation should force writing the data and/or metadata to the underlying storage.
