@@ -17,8 +17,6 @@
  */
 package org.apache.ratis.server.impl;
 
-import java.util.concurrent.TimeoutException;
-import java.util.concurrent.atomic.AtomicLong;
 import org.apache.ratis.client.RaftClient;
 import org.apache.ratis.client.impl.ClientProtoUtils;
 import org.apache.ratis.conf.RaftProperties;
@@ -39,13 +37,13 @@ import org.apache.ratis.protocol.exceptions.TransferLeadershipException;
 import org.apache.ratis.server.DataStreamMap;
 import org.apache.ratis.server.DivisionInfo;
 import org.apache.ratis.server.DivisionProperties;
-import org.apache.ratis.server.leader.FollowerInfo;
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.RaftServerMXBean;
 import org.apache.ratis.server.RaftServerRpc;
 import org.apache.ratis.server.impl.LeaderElection.Phase;
 import org.apache.ratis.server.impl.RetryCacheImpl.CacheEntry;
+import org.apache.ratis.server.leader.FollowerInfo;
 import org.apache.ratis.server.leader.LeaderState;
 import org.apache.ratis.server.leader.LogAppender;
 import org.apache.ratis.server.metrics.LeaderElectionMetrics;
@@ -59,6 +57,7 @@ import org.apache.ratis.server.raftlog.RaftLogIOException;
 import org.apache.ratis.server.storage.RaftStorage;
 import org.apache.ratis.server.storage.RaftStorageDirectory;
 import org.apache.ratis.server.util.ServerStringUtils;
+import org.apache.ratis.statemachine.SnapshotInfo;
 import org.apache.ratis.statemachine.StateMachine;
 import org.apache.ratis.statemachine.TransactionContext;
 import org.apache.ratis.thirdparty.com.google.protobuf.InvalidProtocolBufferException;
@@ -69,6 +68,7 @@ import java.io.File;
 import java.io.IOException;
 import java.nio.file.NoSuchFileException;
 import java.util.*;
+import java.util.concurrent.ExecutorService;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.CompletionException;
 import java.util.concurrent.ExecutionException;
@@ -82,6 +82,7 @@ import java.util.stream.Collectors;
 import static org.apache.ratis.proto.RaftProtos.AppendEntriesReplyProto.AppendResult.INCONSISTENCY;
 import static org.apache.ratis.proto.RaftProtos.AppendEntriesReplyProto.AppendResult.NOT_LEADER;
 import static org.apache.ratis.proto.RaftProtos.AppendEntriesReplyProto.AppendResult.SUCCESS;
+import static org.apache.ratis.server.raftlog.RaftLog.INVALID_LOG_INDEX;
 import static org.apache.ratis.util.LifeCycle.State.EXCEPTION;
 import static org.apache.ratis.util.LifeCycle.State.NEW;
 import static org.apache.ratis.util.LifeCycle.State.PAUSED;
@@ -90,6 +91,7 @@ import static org.apache.ratis.util.LifeCycle.State.RUNNING;
 import static org.apache.ratis.util.LifeCycle.State.STARTING;
 
 import com.codahale.metrics.Timer;
+import org.apache.ratis.util.function.CheckedSupplier;
 
 class RaftServerImpl implements RaftServer.Division,
     RaftServerProtocol, RaftServerAsynchronousProtocol,
@@ -112,6 +114,11 @@ class RaftServerImpl implements RaftServer.Division,
       return isLeader() && getRole().isLeaderReady();
     }
 
+    @Override
+    public RaftPeerId getLeaderId() {
+      return getState().getLeaderId();
+    }
+
     @Override
     public LifeCycle.State getLifeCycleState() {
       return lifeCycle.getCurrentState();
@@ -146,10 +153,8 @@ class RaftServerImpl implements RaftServer.Division,
   private final Info info =  new Info();
 
   private final DivisionProperties divisionProperties;
-  private final int maxTimeoutMs;
   private final TimeDuration leaderStepDownWaitTime;
   private final TimeDuration sleepDeviationThreshold;
-  private final boolean installSnapshotEnabled;
 
   private final LifeCycle lifeCycle;
   private final ServerState state;
@@ -166,10 +171,6 @@ class RaftServerImpl implements RaftServer.Division,
   private final LeaderElectionMetrics leaderElectionMetrics;
   private final RaftServerMetricsImpl raftServerMetrics;
 
-  private final AtomicLong inProgressInstallSnapshotRequest;
-  private final AtomicLong installedSnapshotIndex;
-  private final AtomicBoolean isSnapshotNull;
-
   // To avoid append entry before complete start() method
   // For example, if thread1 start(), but before thread1 startAsFollower(), thread2 receive append entry
   // request, and change state to RUNNING by lifeCycle.compareAndTransition(STARTING, RUNNING),
@@ -178,6 +179,11 @@ class RaftServerImpl implements RaftServer.Division,
   private final AtomicBoolean startComplete;
 
   private final TransferLeadership transferLeadership;
+  private final SnapshotManagementRequestHandler snapshotRequestHandler;
+  private final SnapshotInstallationHandler snapshotInstallationHandler;
+
+  private final ExecutorService serverExecutor;
+  private final ExecutorService clientExecutor;
 
   RaftServerImpl(RaftGroup group, StateMachine stateMachine, RaftServerProxy proxy) throws IOException {
     final RaftPeerId id = proxy.getId();
@@ -188,17 +194,12 @@ class RaftServerImpl implements RaftServer.Division,
 
     final RaftProperties properties = proxy.getProperties();
     this.divisionProperties = new DivisionPropertiesImpl(properties);
-    maxTimeoutMs = properties().maxRpcTimeoutMs();
     leaderStepDownWaitTime = RaftServerConfigKeys.LeaderElection.leaderStepDownWaitTime(properties);
     this.sleepDeviationThreshold = RaftServerConfigKeys.sleepDeviationThreshold(properties);
-    installSnapshotEnabled = RaftServerConfigKeys.Log.Appender.installSnapshotEnabled(properties);
     this.proxy = proxy;
 
     this.state = new ServerState(id, group, properties, this, stateMachine);
     this.retryCache = new RetryCacheImpl(properties);
-    this.inProgressInstallSnapshotRequest = new AtomicLong();
-    this.installedSnapshotIndex = new AtomicLong();
-    this.isSnapshotNull = new AtomicBoolean(false);
     this.dataStreamMap = new DataStreamMapImpl(id);
 
     this.jmxAdapter = new RaftServerJmxAdapter();
@@ -209,15 +210,23 @@ class RaftServerImpl implements RaftServer.Division,
 
     this.startComplete = new AtomicBoolean(false);
 
-    this.raftClient = JavaUtils.memoize(() -> {
-      RaftClient client = RaftClient.newBuilder()
-          .setRaftGroup(group)
-          .setProperties(getRaftServer().getProperties())
-          .build();
-      return client;
-    });
+    this.raftClient = JavaUtils.memoize(() -> RaftClient.newBuilder()
+        .setRaftGroup(group)
+        .setProperties(getRaftServer().getProperties())
+        .build());
 
     this.transferLeadership = new TransferLeadership(this);
+    this.snapshotRequestHandler = new SnapshotManagementRequestHandler(this);
+    this.snapshotInstallationHandler = new SnapshotInstallationHandler(this, properties);
+
+    this.serverExecutor = ConcurrentUtils.newThreadPoolWithMax(
+        RaftServerConfigKeys.ThreadPool.serverCached(properties),
+        RaftServerConfigKeys.ThreadPool.serverSize(properties),
+        id + "-server");
+    this.clientExecutor = ConcurrentUtils.newThreadPoolWithMax(
+        RaftServerConfigKeys.ThreadPool.clientCached(properties),
+        RaftServerConfigKeys.ThreadPool.clientSize(properties),
+        id + "-client");
   }
 
   @Override
@@ -230,7 +239,7 @@ class RaftServerImpl implements RaftServer.Division,
   }
 
   int getMaxTimeoutMs() {
-    return maxTimeoutMs;
+    return properties().maxRpcTimeoutMs();
   }
 
   TimeDuration getRandomElectionTimeout() {
@@ -299,7 +308,10 @@ class RaftServerImpl implements RaftServer.Division,
     final RaftConfigurationImpl conf = getRaftConf();
     if (conf != null && conf.containsInBothConfs(getId())) {
       LOG.info("{}: start as a follower, conf={}", getMemberId(), conf);
-      startAsFollower();
+      startAsPeer(RaftPeerRole.FOLLOWER);
+    } else if (conf != null && conf.containsInConf(getId(), RaftPeerRole.LISTENER)) {
+      LOG.info("{}: start as a listener, conf={}", getMemberId(), conf);
+      startAsPeer(RaftPeerRole.LISTENER);
     } else {
       LOG.info("{}: start with initializing state, conf={}", getMemberId(), conf);
       startInitializing();
@@ -312,8 +324,8 @@ class RaftServerImpl implements RaftServer.Division,
   }
 
   static boolean registerMBean(
-      RaftPeerId id, RaftGroupId groupdId, RaftServerMXBean mBean, JmxRegister jmx) {
-    final String prefix = "Ratis:service=RaftServer,group=" + groupdId + ",id=";
+      RaftPeerId id, RaftGroupId groupId, RaftServerMXBean mBean, JmxRegister jmx) {
+    final String prefix = "Ratis:service=RaftServer,group=" + groupId + ",id=";
     final String registered = jmx.register(mBean, Arrays.asList(
         () -> prefix + id,
         () -> prefix + ObjectName.quote(id.toString())));
@@ -321,11 +333,21 @@ class RaftServerImpl implements RaftServer.Division,
   }
 
   /**
-   * The peer belongs to the current configuration, should start as a follower
+   * The peer belongs to the current configuration, should start as a follower or listener
    */
-  private void startAsFollower() {
-    setRole(RaftPeerRole.FOLLOWER, "startAsFollower");
-    role.startFollowerState(this, "startAsFollower");
+  private void startAsPeer(RaftPeerRole newRole) {
+    Object reason = "";
+    if (newRole == RaftPeerRole.FOLLOWER) {
+      reason = "startAsFollower";
+      setRole(RaftPeerRole.FOLLOWER, reason);
+    } else if (newRole == RaftPeerRole.LISTENER) {
+      reason = "startAsListener";
+      setRole(RaftPeerRole.LISTENER, reason);
+    } else {
+      throw new IllegalArgumentException("Unexpected role " + newRole);
+    }
+    role.startFollowerState(this, reason);
+
     lifeCycle.transition(RUNNING);
   }
 
@@ -449,6 +471,17 @@ class RaftServerImpl implements RaftServer.Division,
       } catch (Exception ignored) {
         LOG.warn("{}: Failed to close raft client", getMemberId(), ignored);
       }
+
+      try {
+        ConcurrentUtils.shutdownAndWait(clientExecutor);
+      } catch (Exception ignored) {
+        LOG.warn(getMemberId() + ": Failed to shutdown clientExecutor", ignored);
+      }
+      try {
+        ConcurrentUtils.shutdownAndWait(serverExecutor);
+      } catch (Exception ignored) {
+        LOG.warn(getMemberId() + ": Failed to shutdown serverExecutor", ignored);
+      }
     });
   }
 
@@ -460,6 +493,9 @@ class RaftServerImpl implements RaftServer.Division,
    */
   private synchronized boolean changeToFollower(long newTerm, boolean force, Object reason) {
     final RaftPeerRole old = role.getCurrentRole();
+    if (old == RaftPeerRole.LISTENER) {
+      throw new IllegalStateException("Unexpected role " + old);
+    }
     final boolean metadataUpdated = state.updateCurrentTerm(newTerm);
 
     if (old != RaftPeerRole.FOLLOWER || force) {
@@ -493,7 +529,8 @@ class RaftServerImpl implements RaftServer.Division,
     getState().setRaftConf(e);
   }
 
-  Collection<CommitInfoProto> getCommitInfos() {
+  @Override
+  public Collection<CommitInfoProto> getCommitInfos() {
     final List<CommitInfoProto> infos = new ArrayList<>();
     // add the commit info of this server
     infos.add(updateCommitInfoCache());
@@ -518,21 +555,6 @@ class RaftServerImpl implements RaftServer.Division,
         getGroup(), getRoleInfoProto(), state.getStorage().getStorageDir().isHealthy());
   }
 
-  private RoleInfoProto getRoleInfoProto(RaftPeer leaderPeerInfo) {
-    RaftPeerRole currentRole = role.getCurrentRole();
-    RoleInfoProto.Builder roleInfo = RoleInfoProto.newBuilder()
-        .setSelf(getPeer().getRaftPeerProto())
-        .setRole(currentRole)
-        .setRoleElapsedTimeMs(role.getRoleElapsedTimeMs());
-    final Optional<FollowerState> fs = role.getFollowerState();
-    final ServerRpcProto leaderInfo =
-        ServerProtoUtils.toServerRpcProto(leaderPeerInfo,
-            fs.map(FollowerState::getLastRpcTime).map(Timestamp::elapsedTimeMs).orElse(0L));
-    roleInfo.setFollowerInfo(FollowerInfoProto.newBuilder().setLeaderInfo(leaderInfo)
-        .setOutstandingOp(fs.map(FollowerState::getOutstandingOp).orElse(0)));
-    return roleInfo.build();
-  }
-
   RoleInfoProto getRoleInfoProto() {
     RaftPeerRole currentRole = role.getCurrentRole();
     RoleInfoProto.Builder roleInfo = RoleInfoProto.newBuilder()
@@ -611,6 +633,13 @@ class RaftServerImpl implements RaftServer.Division,
         .build();
   }
 
+  RaftClientReply newSuccessReply(RaftClientRequest request, long logIndex) {
+    return newReplyBuilder(request)
+        .setSuccess()
+        .setLogIndex(logIndex)
+        .build();
+  }
+
   RaftClientReply newExceptionReply(RaftClientRequest request, RaftException exception) {
     return newReplyBuilder(request)
         .setException(exception)
@@ -667,7 +696,7 @@ class RaftServerImpl implements RaftServer.Division,
     return new NotLeaderException(getMemberId(), conf.getPeer(leaderId), peers);
   }
 
-  private LifeCycle.State assertLifeCycleState(Set<LifeCycle.State> expected) throws ServerNotReadyException {
+  LifeCycle.State assertLifeCycleState(Set<LifeCycle.State> expected) throws ServerNotReadyException {
     return lifeCycle.assertCurrentState((n, c) -> new ServerNotReadyException(
         getMemberId() + " is not in " + expected + ": current state is " + c),
         expected);
@@ -742,6 +771,19 @@ class RaftServerImpl implements RaftServer.Division,
             request.getMessage().getContent().asReadOnlyByteBuffer()));
   }
 
+  <REPLY> CompletableFuture<REPLY> executeSubmitServerRequestAsync(
+      CheckedSupplier<CompletableFuture<REPLY>, IOException> submitFunction) {
+    return CompletableFuture.supplyAsync(
+        () -> JavaUtils.callAsUnchecked(submitFunction, CompletionException::new),
+        serverExecutor).join();
+  }
+
+  CompletableFuture<RaftClientReply> executeSubmitClientRequestAsync(RaftClientRequest request) {
+    return CompletableFuture.supplyAsync(
+        () -> JavaUtils.callAsUnchecked(() -> submitClientRequestAsync(request), CompletionException::new),
+        clientExecutor).join();
+  }
+
   @Override
   public CompletableFuture<RaftClientReply> submitClientRequestAsync(
       RaftClientRequest request) throws IOException {
@@ -902,7 +944,7 @@ class RaftServerImpl implements RaftServer.Division,
     }
   }
 
-  public RaftClientReply transferLeadership(TransferLeadershipRequest request) throws IOException {
+  RaftClientReply transferLeadership(TransferLeadershipRequest request) throws IOException {
     return waitForReply(request, transferLeadershipAsync(request));
   }
 
@@ -921,8 +963,12 @@ class RaftServerImpl implements RaftServer.Division,
     transferLeadership.finish(state.getLeaderId(), false);
   }
 
-  public CompletableFuture<RaftClientReply> transferLeadershipAsync(TransferLeadershipRequest request)
+  CompletableFuture<RaftClientReply> transferLeadershipAsync(TransferLeadershipRequest request)
       throws IOException {
+    if (request.getNewLeader() == null) {
+      return stepDownLeaderAsync(request);
+    }
+
     LOG.info("{}: receive transferLeadership {}", getMemberId(), request);
     assertLifeCycleState(LifeCycle.States.RUNNING);
     assertGroup(request.getRequestorId(), request.getRaftGroupId());
@@ -963,6 +1009,66 @@ class RaftServerImpl implements RaftServer.Division,
     }
   }
 
+  CompletableFuture<RaftClientReply> takeSnapshotAsync(SnapshotManagementRequest request) throws IOException {
+    LOG.info("{}: takeSnapshotAsync {}", getMemberId(), request);
+    assertLifeCycleState(LifeCycle.States.RUNNING);
+    assertGroup(request.getRequestorId(), request.getRaftGroupId());
+
+    //TODO(liuyaolong): get the gap value from shell command
+    long minGapValue = RaftServerConfigKeys.Snapshot.creationGap(proxy.getProperties());
+    final long lastSnapshotIndex = Optional.ofNullable(stateMachine.getLatestSnapshot())
+        .map(SnapshotInfo::getIndex)
+        .orElse(0L);
+    if (state.getLastAppliedIndex() - lastSnapshotIndex < minGapValue) {
+      return CompletableFuture.completedFuture(newSuccessReply(request, lastSnapshotIndex));
+    }
+
+    synchronized (this) {
+      final long installSnapshot = snapshotInstallationHandler.getInProgressInstallSnapshotIndex();
+      // check snapshot install/load
+      if (installSnapshot != INVALID_LOG_INDEX) {
+        String msg = String.format("%s: Failed do snapshot as snapshot (%s) installation is in progress",
+            getMemberId(), installSnapshot);
+        LOG.warn(msg);
+        return CompletableFuture.completedFuture(newExceptionReply(request,new RaftException(msg)));
+      }
+      return snapshotRequestHandler.takingSnapshotAsync(request);
+    }
+  }
+
+  SnapshotManagementRequestHandler getSnapshotRequestHandler() {
+    return snapshotRequestHandler;
+  }
+
+  CompletableFuture<RaftClientReply> leaderElectionManagementAsync(LeaderElectionManagementRequest request)
+      throws IOException {
+    LOG.info("{} receive leaderElectionManagement request {}", getMemberId(), request);
+    assertLifeCycleState(LifeCycle.States.RUNNING);
+    assertGroup(request.getRequestorId(), request.getRaftGroupId());
+
+    final LeaderElectionManagementRequest.Pause pause = request.getPause();
+    if (pause != null) {
+      getRole().setLeaderElectionPause(true);
+      return CompletableFuture.completedFuture(newSuccessReply(request));
+    }
+    final LeaderElectionManagementRequest.Resume resume = request.getResume();
+    if (resume != null) {
+      getRole().setLeaderElectionPause(false);
+      return CompletableFuture.completedFuture(newSuccessReply(request));
+    }
+    return JavaUtils.completeExceptionally(new UnsupportedOperationException(
+        getId() + ": Request not supported " + request));
+  }
+
+  CompletableFuture<RaftClientReply> stepDownLeaderAsync(TransferLeadershipRequest request) throws IOException {
+    LOG.info("{} receive stepDown leader request {}", getMemberId(), request);
+    assertLifeCycleState(LifeCycle.States.RUNNING);
+    assertGroup(request.getRequestorId(), request.getRaftGroupId());
+
+    return role.getLeaderState().map(leader -> leader.submitStepDownRequestAsync(request))
+        .orElseGet(() -> CompletableFuture.completedFuture(newSuccessReply(request)));
+  }
+
   public RaftClientReply setConfiguration(SetConfigurationRequest request) throws IOException {
     return waitForReply(request, setConfigurationAsync(request));
   }
@@ -1154,7 +1260,7 @@ class RaftServerImpl implements RaftServer.Division,
     }
   }
 
-  private Optional<FollowerState> updateLastRpcTime(FollowerState.UpdateType updateType) {
+  Optional<FollowerState> updateLastRpcTime(FollowerState.UpdateType updateType) {
     final Optional<FollowerState> fs = role.getFollowerState();
     if (fs.isPresent() && lifeCycle.getCurrentState() == RUNNING) {
       fs.get().updateLastRpcTime(updateType);
@@ -1209,7 +1315,7 @@ class RaftServerImpl implements RaftServer.Division,
       if (!recognized) {
         final AppendEntriesReplyProto reply = ServerProtoUtils.toAppendEntriesReplyProto(
             leaderId, getMemberId(), currentTerm, followerCommit, state.getNextIndex(), NOT_LEADER, callId,
-            RaftLog.INVALID_LOG_INDEX, isHeartbeat);
+            INVALID_LOG_INDEX, isHeartbeat);
         if (LOG.isDebugEnabled()) {
           LOG.debug("{}: Not recognize {} (term={}) as leader, state: {} reply: {}",
               getMemberId(), leaderId, leaderTerm, state, ServerStringUtils.toAppendEntriesReplyString(reply));
@@ -1263,7 +1369,7 @@ class RaftServerImpl implements RaftServer.Division,
         updateCommitInfoCache();
         final long n = isHeartbeat? state.getLog().getNextIndex(): entries[entries.length - 1].getIndex() + 1;
         final long matchIndex = entries.length != 0 ? entries[entries.length - 1].getIndex() :
-            RaftLog.INVALID_LOG_INDEX;
+            INVALID_LOG_INDEX;
         reply = ServerProtoUtils.toAppendEntriesReplyProto(leaderId, getMemberId(), currentTerm,
             state.getLog().getLastCommittedIndex(), n, SUCCESS, callId, matchIndex,
             isHeartbeat);
@@ -1284,35 +1390,31 @@ class RaftServerImpl implements RaftServer.Division,
 
     final AppendEntriesReplyProto reply = ServerProtoUtils.toAppendEntriesReplyProto(
         leaderId, getMemberId(), currentTerm, followerCommit, replyNextIndex, INCONSISTENCY, callId,
-        RaftLog.INVALID_LOG_INDEX, isHeartbeat);
+        INVALID_LOG_INDEX, isHeartbeat);
     LOG.info("{}: inconsistency entries. Reply:{}", getMemberId(), ServerStringUtils.toAppendEntriesReplyString(reply));
     return reply;
   }
 
   private long checkInconsistentAppendEntries(TermIndex previous, LogEntryProto... entries) {
     // Check if a snapshot installation through state machine is in progress.
-    final long installSnapshot = inProgressInstallSnapshotRequest.get();
-    if (installSnapshot != 0) {
+    final long installSnapshot = snapshotInstallationHandler.getInProgressInstallSnapshotIndex();
+    if (installSnapshot != INVALID_LOG_INDEX) {
       LOG.info("{}: Failed appendEntries as snapshot ({}) installation is in progress", getMemberId(), installSnapshot);
       return state.getNextIndex();
     }
 
-    // Check that the first log entry is greater than the snapshot index in the latest snapshot.
-    // If not, reply to the leader the new next index.
+    // Check that the first log entry is greater than the snapshot index in the latest snapshot and follower's last
+    // committed index. If not, reply to the leader the new next index.
     if (entries != null && entries.length > 0) {
       final long firstEntryIndex = entries[0].getIndex();
       final long snapshotIndex = state.getSnapshotIndex();
-      if (snapshotIndex > 0 && snapshotIndex >= firstEntryIndex) {
-        LOG.info("{}: Failed appendEntries: the first entry (index {}) is already in snapshot (snapshot index: {})",
-            getMemberId(), firstEntryIndex, snapshotIndex);
-        return snapshotIndex + 1;
-      }
-
       final long commitIndex =  state.getLog().getLastCommittedIndex();
-      if (commitIndex > 0 && commitIndex >= firstEntryIndex) {
-        LOG.info("{}: Failed appendEntries: the first entry (index {}) is already committed (commit index: {})",
-            getMemberId(), firstEntryIndex, commitIndex);
-        return commitIndex + 1;
+      final long nextIndex = Math.max(snapshotIndex, commitIndex);
+      if (nextIndex > INVALID_LOG_INDEX && nextIndex >= firstEntryIndex) {
+        LOG.info("{}: Failed appendEntries as the first entry (index {})" +
+                " already exists (snapshotIndex: {}, commitIndex: {})",
+            getMemberId(), firstEntryIndex, snapshotIndex, commitIndex);
+        return nextIndex + 1;
       }
     }
 
@@ -1328,25 +1430,10 @@ class RaftServerImpl implements RaftServer.Division,
 
   @Override
   public InstallSnapshotReplyProto installSnapshot(InstallSnapshotRequestProto request) throws IOException {
-    if (LOG.isInfoEnabled()) {
-      LOG.info("{}: receive installSnapshot: {}", getMemberId(),
-          ServerStringUtils.toInstallSnapshotRequestString(request));
-    }
-    final InstallSnapshotReplyProto reply;
-    try {
-      reply = installSnapshotImpl(request);
-    } catch (Exception e) {
-      LOG.error("{}: installSnapshot failed", getMemberId(), e);
-      throw e;
-    }
-    if (LOG.isInfoEnabled()) {
-      LOG.info("{}: reply installSnapshot: {}", getMemberId(),
-          ServerStringUtils.toInstallSnapshotReplyString(reply));
-    }
-    return reply;
+    return snapshotInstallationHandler.installSnapshot(request);
   }
 
-  boolean pause() throws IOException {
+  boolean pause() {
     // TODO: should pause() be limited on only working for a follower?
 
     // Now the state of lifeCycle should be PAUSING, which will prevent future other operations.
@@ -1372,7 +1459,7 @@ class RaftServerImpl implements RaftServer.Division,
       try {
         stateMachine.reinitialize();
       } catch (IOException e) {
-        LOG.warn("Failed to reinitialize statemachine: {}", stateMachine.toString());
+        LOG.warn("Failed to reinitialize statemachine: {}", stateMachine);
         lifeCycle.compareAndTransition(STARTING, EXCEPTION);
         throw e;
       }
@@ -1430,228 +1517,6 @@ class RaftServerImpl implements RaftServer.Division,
     }
   }
 
-  private InstallSnapshotReplyProto installSnapshotImpl(InstallSnapshotRequestProto request) throws IOException {
-    final RaftRpcRequestProto r = request.getServerRequest();
-    final RaftPeerId leaderId = RaftPeerId.valueOf(r.getRequestorId());
-    final RaftGroupId leaderGroupId = ProtoUtils.toRaftGroupId(r.getRaftGroupId());
-    CodeInjectionForTesting.execute(INSTALL_SNAPSHOT, getId(),
-        leaderId, request);
-
-    assertLifeCycleState(LifeCycle.States.STARTING_OR_RUNNING);
-    assertGroup(leaderId, leaderGroupId);
-
-    InstallSnapshotReplyProto reply = null;
-    // Check if install snapshot from Leader is enabled
-    if (installSnapshotEnabled) {
-      // Leader has sent InstallSnapshot request with SnapshotInfo. Install the snapshot.
-      if (request.hasSnapshotChunk()) {
-        reply = checkAndInstallSnapshot(request, leaderId);
-      }
-    } else {
-      // Leader has only sent a notification to install snapshot. Inform State Machine to install snapshot.
-      if (request.hasNotification()) {
-        reply = notifyStateMachineToInstallSnapshot(request, leaderId);
-      }
-    }
-
-    if (reply != null) {
-      if (request.hasLastRaftConfigurationLogEntryProto()) {
-        // Set the configuration included in the snapshot
-        LogEntryProto newConfLogEntryProto =
-            request.getLastRaftConfigurationLogEntryProto();
-        LOG.info("{}: set new configuration {} from snapshot", getMemberId(),
-            newConfLogEntryProto);
-
-        state.setRaftConf(newConfLogEntryProto);
-        state.writeRaftConfiguration(newConfLogEntryProto);
-        stateMachine.event().notifyConfigurationChanged(newConfLogEntryProto.getTerm(), newConfLogEntryProto.getIndex(),
-            newConfLogEntryProto.getConfigurationEntry());
-      }
-      return reply;
-    }
-
-    // There is a mismatch between configurations on leader and follower.
-    final InstallSnapshotReplyProto failedReply = ServerProtoUtils.toInstallSnapshotReplyProto(
-        leaderId, getMemberId(), InstallSnapshotResult.CONF_MISMATCH);
-    LOG.error("{}: Configuration Mismatch ({}): Leader {} has it set to {} but follower {} has it set to {}",
-        getMemberId(), RaftServerConfigKeys.Log.Appender.INSTALL_SNAPSHOT_ENABLED_KEY,
-        leaderId, request.hasSnapshotChunk(), getId(), installSnapshotEnabled);
-    return failedReply;
-  }
-
-  private InstallSnapshotReplyProto checkAndInstallSnapshot(
-      InstallSnapshotRequestProto request, RaftPeerId leaderId) throws IOException {
-    final long currentTerm;
-    final long leaderTerm = request.getLeaderTerm();
-    InstallSnapshotRequestProto.SnapshotChunkProto snapshotChunkRequest = request.getSnapshotChunk();
-    final long lastIncludedIndex = snapshotChunkRequest.getTermIndex().getIndex();
-    synchronized (this) {
-      final boolean recognized = state.recognizeLeader(leaderId, leaderTerm);
-      currentTerm = state.getCurrentTerm();
-      if (!recognized) {
-        final InstallSnapshotReplyProto reply = ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(),
-            currentTerm, snapshotChunkRequest.getRequestIndex(), InstallSnapshotResult.NOT_LEADER);
-        LOG.warn("{}: Failed to recognize leader for installSnapshot chunk.", getMemberId());
-        return reply;
-      }
-      changeToFollowerAndPersistMetadata(leaderTerm, "installSnapshot");
-      state.setLeader(leaderId, "installSnapshot");
-
-      updateLastRpcTime(FollowerState.UpdateType.INSTALL_SNAPSHOT_START);
-      try {
-        // Check and append the snapshot chunk. We simply put this in lock
-        // considering a follower peer requiring a snapshot installation does not
-        // have a lot of requests
-        Preconditions.assertTrue(
-            state.getLog().getNextIndex() <= lastIncludedIndex,
-            "%s log's next id is %s, last included index in snapshot is %s",
-            getMemberId(), state.getLog().getNextIndex(), lastIncludedIndex);
-
-        //TODO: We should only update State with installed snapshot once the request is done.
-        state.installSnapshot(request);
-
-        // update the committed index
-        // re-load the state machine if this is the last chunk
-        if (snapshotChunkRequest.getDone()) {
-          state.reloadStateMachine(lastIncludedIndex);
-        }
-      } finally {
-        updateLastRpcTime(FollowerState.UpdateType.INSTALL_SNAPSHOT_COMPLETE);
-      }
-    }
-    if (snapshotChunkRequest.getDone()) {
-      LOG.info("{}: successfully install the entire snapshot-{}", getMemberId(), lastIncludedIndex);
-    }
-    return ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(),
-        currentTerm, snapshotChunkRequest.getRequestIndex(), InstallSnapshotResult.SUCCESS);
-  }
-
-  private InstallSnapshotReplyProto notifyStateMachineToInstallSnapshot(
-      InstallSnapshotRequestProto request, RaftPeerId leaderId) throws IOException {
-    final long currentTerm;
-    final long leaderTerm = request.getLeaderTerm();
-    final TermIndex firstAvailableLogTermIndex = TermIndex.valueOf(
-        request.getNotification().getFirstAvailableTermIndex());
-    final long firstAvailableLogIndex = firstAvailableLogTermIndex.getIndex();
-    synchronized (this) {
-      final boolean recognized = state.recognizeLeader(leaderId, leaderTerm);
-      currentTerm = state.getCurrentTerm();
-      if (!recognized) {
-        final InstallSnapshotReplyProto reply = ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(),
-            currentTerm, InstallSnapshotResult.NOT_LEADER, -1);
-        LOG.warn("{}: Failed to recognize leader for installSnapshot notification.", getMemberId());
-        return reply;
-      }
-      changeToFollowerAndPersistMetadata(leaderTerm, "installSnapshot");
-      state.setLeader(leaderId, "installSnapshot");
-      long snapshotIndex = state.getSnapshotIndex();
-
-      updateLastRpcTime(FollowerState.UpdateType.INSTALL_SNAPSHOT_NOTIFICATION);
-      if (inProgressInstallSnapshotRequest.compareAndSet(0, firstAvailableLogIndex)) {
-        LOG.info("{}: Received notification to install snapshot at index {}", getMemberId(), firstAvailableLogIndex);
-        // Check if snapshot index is already at par or ahead of the first
-        // available log index of the Leader.
-        if (snapshotIndex + 1 >= firstAvailableLogIndex && firstAvailableLogIndex > 0) {
-          // State Machine has already installed the snapshot. Return the
-          // latest snapshot index to the Leader.
-
-          inProgressInstallSnapshotRequest.compareAndSet(firstAvailableLogIndex, 0);
-          LOG.info("{}: InstallSnapshot notification result: {}, current snapshot index: {}", getMemberId(),
-              InstallSnapshotResult.ALREADY_INSTALLED, snapshotIndex);
-          return ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(), currentTerm,
-              InstallSnapshotResult.ALREADY_INSTALLED, snapshotIndex);
-        }
-
-        Optional<RaftPeerProto> leaderPeerInfo = null;
-        if (request.hasLastRaftConfigurationLogEntryProto()) {
-          List<RaftPeerProto> peerList = request.getLastRaftConfigurationLogEntryProto().getConfigurationEntry()
-              .getPeersList();
-          leaderPeerInfo = peerList.stream().filter(p -> RaftPeerId.valueOf(p.getId()).equals(leaderId)).findFirst();
-          Preconditions.assertTrue(leaderPeerInfo.isPresent());
-        }
-
-        // For the cases where RaftConf is empty on newly started peer with
-        // empty peer list, we retrieve leader info from
-        // installSnapShotRequestProto.
-        RoleInfoProto roleInfoProto =
-            getRaftConf().getPeer(state.getLeaderId()) == null ?
-                getRoleInfoProto(ProtoUtils.toRaftPeer(leaderPeerInfo.get())) :
-                getRoleInfoProto();
-        // This is the first installSnapshot notify request for this term and
-        // index. Notify the state machine to install the snapshot.
-        LOG.info("{}: notifyInstallSnapshot: nextIndex is {} but the leader's first available index is {}.",
-            getMemberId(), state.getLog().getNextIndex(), firstAvailableLogIndex);
-        try {
-          stateMachine.followerEvent().notifyInstallSnapshotFromLeader(roleInfoProto, firstAvailableLogTermIndex)
-              .whenComplete((reply, exception) -> {
-                if (exception != null) {
-                  LOG.warn("{}: Failed to notify StateMachine to InstallSnapshot. Exception: {}",
-                      getMemberId(), exception.getMessage());
-                  inProgressInstallSnapshotRequest.compareAndSet(firstAvailableLogIndex, 0);
-                  return;
-                }
-
-                if (reply != null) {
-                  LOG.info("{}: StateMachine successfully installed snapshot index {}. Reloading the StateMachine.",
-                      getMemberId(), reply.getIndex());
-                  stateMachine.pause();
-                  state.updateInstalledSnapshotIndex(reply);
-                  state.reloadStateMachine(reply.getIndex());
-                  installedSnapshotIndex.set(reply.getIndex());
-                } else {
-                  isSnapshotNull.set(true);
-                  if (LOG.isDebugEnabled()) {
-                    LOG.debug("{}: StateMachine could not install snapshot as it is not available", this);
-                  }
-                }
-              // wait for 10 seconds for statemachine to install snapshot
-              }).get(1, TimeUnit.SECONDS);
-        } catch (ExecutionException t) {
-          inProgressInstallSnapshotRequest.compareAndSet(firstAvailableLogIndex, 0);
-          LOG.warn("{}: Failed to notify StateMachine to InstallSnapshot. ExecutionException: {}",
-              getMemberId(), t.getMessage());
-          throw new IOException("Failed to install snapshot");
-        } catch (InterruptedException | TimeoutException t) {}
-
-        if (LOG.isDebugEnabled()) {
-          LOG.debug("{}: StateMachine is processing Snapshot Installation Request.", getMemberId());
-        }
-      } else {
-        if (LOG.isDebugEnabled()) {
-          LOG.debug("{}: StateMachine is already installing a snapshot.", getMemberId());
-        }
-      }
-
-      // If the snapshot is null or unavailable, return SNAPSHOT_UNAVAILABLE.
-      if (isSnapshotNull.compareAndSet(true, false)) {
-        LOG.info("{}: InstallSnapshot notification result: {}", getMemberId(),
-            InstallSnapshotResult.SNAPSHOT_UNAVAILABLE);
-        inProgressInstallSnapshotRequest.compareAndSet(firstAvailableLogIndex, 0);
-        return ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(),
-            currentTerm, InstallSnapshotResult.SNAPSHOT_UNAVAILABLE, -1);
-      }
-
-      // If a snapshot has been installed, return SNAPSHOT_INSTALLED with the installed snapshot index and reset
-      // installedSnapshotIndex to 0.
-      long latestInstalledSnapshotIndex = this.installedSnapshotIndex.getAndSet(0);
-      if (latestInstalledSnapshotIndex > 0) {
-        LOG.info("{}: InstallSnapshot notification result: {}, at index: {}", getMemberId(),
-            InstallSnapshotResult.SNAPSHOT_INSTALLED, latestInstalledSnapshotIndex);
-        inProgressInstallSnapshotRequest.compareAndSet(firstAvailableLogIndex, 0);
-        return ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(),
-            currentTerm, InstallSnapshotResult.SNAPSHOT_INSTALLED, latestInstalledSnapshotIndex);
-      }
-
-      // Otherwise, Snapshot installation is in progress.
-      if (LOG.isDebugEnabled()) {
-        LOG.debug("{}: InstallSnapshot notification result: {}", getMemberId(),
-            InstallSnapshotResult.IN_PROGRESS);
-      }
-      return ServerProtoUtils.toInstallSnapshotReplyProto(leaderId, getMemberId(),
-          currentTerm, InstallSnapshotResult.IN_PROGRESS, -1);
-    }
-  }
-
   void submitUpdateCommitEvent() {
     role.getLeaderState().ifPresent(LeaderStateImpl::submitUpdateCommitEvent);
   }
