@@ -27,7 +27,6 @@ import static org.apache.ratis.server.metrics.RaftServerMetricsImpl.REQUEST_BYTE
 import static org.apache.ratis.server.metrics.RaftServerMetricsImpl.RESOURCE_LIMIT_HIT_COUNTER;
 
 import com.codahale.metrics.Gauge;
-import org.apache.commons.lang3.RandomStringUtils;
 import org.apache.log4j.Level;
 import org.apache.ratis.BaseTest;
 import org.apache.ratis.protocol.RaftGroup;
@@ -40,7 +39,6 @@ import org.apache.ratis.client.RaftClientRpc;
 import org.apache.ratis.client.impl.RaftClientTestUtil;
 import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.grpc.client.GrpcClientProtocolClient;
-import org.apache.ratis.grpc.client.GrpcClientProtocolService;
 import org.apache.ratis.proto.RaftProtos;
 import org.apache.ratis.proto.RaftProtos.RaftPeerRole;
 import org.apache.ratis.protocol.exceptions.AlreadyClosedException;
@@ -56,6 +54,7 @@ import org.apache.ratis.server.impl.RaftServerTestUtil;
 import org.apache.ratis.server.metrics.RaftServerMetricsImpl;
 import org.apache.ratis.statemachine.SimpleStateMachine4Testing;
 import org.apache.ratis.statemachine.StateMachine;
+import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.ProtoUtils;
 import org.apache.ratis.util.SizeInBytes;
@@ -72,12 +71,12 @@ import java.util.List;
 import java.util.SortedMap;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.ExecutionException;
+import java.util.concurrent.ThreadLocalRandom;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicLong;
 
 public class TestRaftServerWithGrpc extends BaseTest implements MiniRaftClusterWithGrpc.FactoryGet {
   {
-    Log4jUtils.setLogLevel(GrpcClientProtocolService.LOG, Level.ALL);
     Log4jUtils.setLogLevel(GrpcClientProtocolClient.LOG, Level.ALL);
   }
 
@@ -230,7 +229,6 @@ public class TestRaftServerWithGrpc extends BaseTest implements MiniRaftClusterW
     SimpleStateMachine4Testing stateMachine = SimpleStateMachine4Testing.get(cluster.getLeader());
     stateMachine.blockFlushStateMachineData();
 
-    String message = "2nd Message";
     // Block stateMachine flush data, so that 2nd request will not be
     // completed, and so it will not be removed from pending request map.
     List<RaftClient> clients = new ArrayList<>();
@@ -238,7 +236,7 @@ public class TestRaftServerWithGrpc extends BaseTest implements MiniRaftClusterW
     try {
       RaftClient client = cluster.createClient(cluster.getLeader().getId(), RetryPolicies.noRetry());
       clients.add(client);
-      client.async().send(new SimpleMessage(message));
+      client.async().send(new SimpleMessage("2nd Message"));
 
 
       final SortedMap<String, Gauge> gaugeMap = getRaftServerMetrics(cluster.getLeader())
@@ -249,7 +247,7 @@ public class TestRaftServerWithGrpc extends BaseTest implements MiniRaftClusterW
       for (int i = 0; i < 10; i++) {
         client = cluster.createClient(cluster.getLeader().getId(), RetryPolicies.noRetry());
         clients.add(client);
-        client.async().send(new SimpleMessage(message));
+        client.async().send(new SimpleMessage("message " + i));
       }
 
       // Because we have passed 11 requests, and the element queue size is 10.
@@ -262,8 +260,10 @@ public class TestRaftServerWithGrpc extends BaseTest implements MiniRaftClusterW
       // and byte size counter limit will be hit.
 
       client = cluster.createClient(cluster.getLeader().getId(), RetryPolicies.noRetry());
-      client.async().send(new SimpleMessage(RandomStringUtils
-          .random(SizeInBytes.valueOf("1025kb").getSizeInt(), true, false)));
+      final SizeInBytes size = SizeInBytes.valueOf("1025kb");
+      final ByteString bytes = randomByteString(size.getSizeInt());
+      Assert.assertEquals(size.getSizeInt(), bytes.size());
+      client.async().send(new SimpleMessage(size + "-message", bytes));
       clients.add(client);
 
       RaftTestUtil.waitFor(() -> getRaftServerMetrics(cluster.getLeader())
@@ -277,6 +277,19 @@ public class TestRaftServerWithGrpc extends BaseTest implements MiniRaftClusterW
       }
     }
   }
+  
+  static ByteString randomByteString(int size) {
+    final ByteString.Output out = ByteString.newOutput(size);
+    final ThreadLocalRandom random = ThreadLocalRandom.current();
+    final byte[] buffer = new byte[4096];
+    for(; size > 0; ) {
+      random.nextBytes(buffer);
+      final int n = Math.min(size, buffer.length);
+      out.write(buffer, 0, n);
+      size -= n;
+    }
+    return out.toByteString();
+  }
 
   static RaftServerMetricsImpl getRaftServerMetrics(RaftServer.Division division) {
     return (RaftServerMetricsImpl) division.getRaftServerMetrics();
