@@ -21,7 +21,6 @@ import org.apache.log4j.Level;
 import org.apache.ratis.WatchRequestTests;
 import org.apache.ratis.client.RaftClient;
 import org.apache.ratis.client.impl.UnorderedAsync;
-import org.apache.ratis.grpc.client.GrpcClientProtocolService;
 import org.apache.ratis.grpc.client.GrpcClientRpc;
 import org.apache.ratis.util.Log4jUtils;
 
@@ -29,7 +28,6 @@ public class TestWatchRequestWithGrpc
     extends WatchRequestTests<MiniRaftClusterWithGrpc>
     implements MiniRaftClusterWithGrpc.FactoryGet {
   {
-    Log4jUtils.setLogLevel(GrpcClientProtocolService.LOG, Level.ALL);
     Log4jUtils.setLogLevel(GrpcClientRpc.LOG, Level.ALL);
     Log4jUtils.setLogLevel(UnorderedAsync.LOG, Level.ALL);
     Log4jUtils.setLogLevel(RaftClient.LOG, Level.ALL);
